/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* These are simple XML parsing routines, used by loaduser and other
 * programs.  This is not a general XML parser, because dump files are
 * not really XML.
 */

#include "common.h"
#include "str.h"
#include "vstr.h"
#include "xmlparse.h"

#include <string.h>
#include <ctype.h>

/* XMLUNQUOTE - Expand &quot to " and so forth.  This is done in place.
 */

void xmlunquote(char *s)
{
    char *d;

    for (d= s; *s != '\0'; *s++)
    {
	if (*s == '&')
	{
	    if (!strncmp(s+1, "lt;", 3))
	    {	*(d++)= '<'; s+= 3; }
	    else if (!strncmp(s+1, "gt;", 3))
	    {	*(d++)= '>'; s+= 3; }
	    else if (!strncmp(s+1, "amp;", 4))
	    {	*(d++)= '&'; s+= 4; }
	    else if (!strncmp(s+1, "quot;", 5))
	    {	*(d++)= '"'; s+= 5; }
	    else if (!strncmp(s+1, "apos;", 5))
	    {	*(d++)= '\''; s+= 5; }
	    else
		*(d++)= *s;
	}
	else
	    *(d++)= *s;
    }
    *d= '\0';
}


/* GETATTR - Read an attribute.  This is meant to be called after getinput()
 * returns START_ELEMENT.  On each call it returns an attribute and a value,
 * both in malloced memory.  Returns false if no more attributes exist.
 */

int getattr(FILE *fp, char **var, char **val)
{
    VSTRING in;
    int ch;

    /* Skip white space */
    while (isascii(ch= fgetc(fp)) && isspace(ch))
	;

    if (ch == '/') { ungetc(ch,fp); return FALSE; }
    if (ch == EOF || ch == '>') return FALSE;

    if (!isascii(ch) || !isalnum(ch))
	die("Bad File Format");

    /* read attrbute name */
    vs_new(&in,20);
    for ( ; isascii(ch) && isalnum(ch); ch= fgetc(fp))
	*vs_inc(&in)= ch;

    if (!isascii(ch))
	die("Bad File Format");

    if (ch == '=')
    {
	*var= vs_return(&in);
	vs_new(&in,20);
	ch= fgetc(fp);
	if (ch == '"' || ch == '\'')
	{
	    /* Read quoted string */
	    int quote= ch;
	    while ((ch= fgetc(fp)) != EOF && ch != quote)
		*vs_inc(&in)= ch;
	}
	else
	{
	    /* Read unquoted string */
	    for ( ; isascii(ch) && ch != '>' && ch != '/' && !isspace(ch);
		    ch= fgetc(fp))
		*vs_inc(&in)= ch;
	    ungetc(ch,fp);
	}

	*val= vs_return(&in);
	return TRUE;
    }
    else if (ch == EOF || ch == '>' || ch == '/' || isspace(ch))
    {
	if (ch == '/') ungetc(ch,fp);
	*var= vs_return(&in);
	*val= NULL;
	return TRUE;
    }
    else
	die ("Bad File Format");
}


/* GETINPUT - Read some input.  Returns one of the following codes:
 *    START_ELEMENT:  Got a <something> element.  Data points to element name
 *        string.  After getting this, should call getattr() until it fails.
 *    END_ELEMENT:  Got a </something> element.  Data points to element name.
 *        If data is NULL, then we hit a /> at the end of a tag, and the
 *        immediately preceeding START_ELEMENT has ended.
 *    CONTENT:  Got some lines of data.  Data points to string.
 * Values pointed to by data are in dynamic storage to be freed by caller
 * if not NULL.
 */

int getinput(FILE *fp, char **data)
{
    int ch;
    int type;
    VSTRING in;

    *data= NULL;

    /* Skip white space at start of line.  Also skips blank lines. */
    while (isascii(ch= fgetc(fp)) && isspace(ch))
	;

    /* Recognize a '/>' ending to a tag */
    if (ch == '/')
    {
	ch= fgetc(fp);
	if (ch != '>')
	    die("Bad File Format");
	return END_ELEMENT;
    }

    if (ch == '<')
    {
	vs_new(&in,20);

	if ((ch= fgetc(fp)) == '/')
	{
	    type= END_ELEMENT;
	    ch= fgetc(fp);
	}
	else
	    type= START_ELEMENT;

	for ( ; isascii(ch) && isalnum(ch); ch= fgetc(fp))
	    *vs_inc(&in)= islower(ch) ? toupper(ch) : ch;

	if (type == END_ELEMENT)
	{
	    /* On end element, skip to '>' */
	    for ( ; ch != EOF && ch != '>'; ch= fgetc(fp))
		*vs_inc(&in)= islower(ch) ? toupper(ch) : ch;
	}
	else
	    ungetc(ch,fp);

	*data= vs_return(&in);
    }
    else if (ch == EOF)
	type= END_OF_DATA;
    else
    {
	int newline= TRUE;
	type= CONTENT;
	vs_new(&in,0);

	for ( ; ch != EOF && (!newline || ch != '<'); ch= fgetc(fp))
	{
	    *vs_inc(&in)= ch;
	    if (ch == '\n')
		newline= TRUE;
	    else if (!isascii(ch) || !isspace(ch))
		newline= FALSE;
	}
	if (ch != EOF) ungetc(ch,fp);

	*data= vs_return(&in);
	xmlunquote(*data);
    }
    return type;
}
